package linux

import (
	"bufio"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"time"

	"github.com/abrander/agento"
)

var previous *CpuStats

type CpuStats struct {
	sampletime       time.Time                 `json:"-"`
	Cpu              map[string]*SingleCpuStat `json:"cpu"`
	Interrupts       float64                   `json:"in"`
	ContextSwitches  float64                   `json:"ct"`
	Forks            float64                   `json:"pr"`
	RunningProcesses int64                     `json:"ru"` // Since 2.5.45
	BlockedProcesses int64                     `json:"bl"` // Since 2.5.45
}

func GetCpuStats() *CpuStats {
	stat := CpuStats{}
	stat.Cpu = make(map[string]*SingleCpuStat)

	path := filepath.Join("/proc/stat")
	file, err := os.Open(path)
	if err != nil {
		return &stat
	}
	defer file.Close()

	stat.sampletime = time.Now()
	scanner := bufio.NewScanner(file)
	for scanner.Scan() {
		text := scanner.Text()

		data := strings.Fields(strings.Trim(text, " "))
		if len(data) < 2 {
			continue
		}

		// cpu* lines
		if strings.HasPrefix(data[0], "cpu") {
			s := SingleCpuStat{}
			s.ReadArray(data)

			key := "cpu." + data[0][3:]

			if data[0] == "cpu" {
				key = data[0]
			}

			stat.Cpu[key] = &s
		}

		switch data[0] {
		case "intr":
			stat.Interrupts, _ = strconv.ParseFloat(data[1], 64)
		case "ctxt":
			stat.ContextSwitches, _ = strconv.ParseFloat(data[1], 64)
		case "processes":
			stat.Forks, _ = strconv.ParseFloat(data[1], 64)
		case "procs_running":
			stat.RunningProcesses, _ = strconv.ParseInt(data[1], 10, 64)
		case "procs_blocked":
			stat.BlockedProcesses, _ = strconv.ParseInt(data[1], 10, 64)
		}
	}

	ret := stat.Sub(previous)
	previous = &stat

	return ret
}

func (c *CpuStats) Sub(previous *CpuStats) *CpuStats {
	if previous == nil {
		return &CpuStats{}
	}

	diff := CpuStats{}
	diff.Cpu = make(map[string]*SingleCpuStat)

	duration := float64(c.sampletime.Sub(previous.sampletime)) / float64(time.Second)
	for key, value := range c.Cpu {
		diff.Cpu[key] = value.Sub(previous.Cpu[key], duration)
	}

	diff.Interrupts = Round((c.Interrupts-previous.Interrupts)/duration, 1)
	diff.ContextSwitches = Round((c.ContextSwitches-previous.ContextSwitches)/duration, 1)
	diff.Forks = Round((c.Forks-previous.Forks)/duration, 1)

	// These are not accumulated
	diff.RunningProcesses = c.RunningProcesses
	diff.BlockedProcesses = c.BlockedProcesses

	return &diff
}

type SingleCpuStat struct {
	User      float64
	Nice      float64
	System    float64
	Idle      float64
	IoWait    float64 // Since 2.5.41
	Irq       float64 // Since 2.6.0-test4
	SoftIrq   float64 // Since 2.6.0-test4
	Steal     float64 // Since 2.6.11
	Guest     float64 // Since 2.6.24
	GuestNice float64 // Since 2.6.33
}

func (s *SingleCpuStat) ReadArray(data []string) {
	l := len(data)

	if l > 4 {
		s.User, _ = strconv.ParseFloat(data[1], 64)
		s.Nice, _ = strconv.ParseFloat(data[2], 64)
		s.System, _ = strconv.ParseFloat(data[3], 64)
		s.Idle, _ = strconv.ParseFloat(data[4], 64)
	}

	if l > 5 {
		s.IoWait, _ = strconv.ParseFloat(data[5], 64)
	}

	if l > 6 {
		s.Irq, _ = strconv.ParseFloat(data[6], 64)
	}

	if l > 7 {
		s.SoftIrq, _ = strconv.ParseFloat(data[7], 64)
	}

	if l > 8 {
		s.Steal, _ = strconv.ParseFloat(data[8], 64)
	}

	if l > 9 {
		s.Guest, _ = strconv.ParseFloat(data[9], 64)
	}

	if l > 10 {
		s.GuestNice, _ = strconv.ParseFloat(data[10], 64)
	}
}

func (s SingleCpuStat) MarshalJSON() ([]byte, error) {
	var a [10]float64

	a[0] = Round(s.User, 1)
	a[1] = Round(s.Nice, 1)
	a[2] = Round(s.System, 1)
	a[3] = Round(s.Idle, 1)
	a[4] = Round(s.IoWait, 1)
	a[5] = Round(s.Irq, 1)
	a[6] = Round(s.SoftIrq, 1)
	a[7] = Round(s.Steal, 1)
	a[8] = Round(s.Guest, 1)
	a[9] = Round(s.GuestNice, 1)

	return json.Marshal(a)
}

func (s *SingleCpuStat) UnmarshalJSON(b []byte) error {
	var a [10]float64

	err := json.Unmarshal(b, &a)
	if err != nil {
		return err
	}

	s.User = a[0]
	s.Nice = a[1]
	s.System = a[2]
	s.Idle = a[3]
	s.IoWait = a[4]
	s.Irq = a[5]
	s.SoftIrq = a[6]
	s.Steal = a[7]
	s.Guest = a[8]
	s.GuestNice = a[9]

	return err
}

func (s *SingleCpuStat) Sub(previous *SingleCpuStat, factor float64) *SingleCpuStat {
	diff := SingleCpuStat{}

	diff.User = (s.User - previous.User) / factor
	diff.Nice = (s.Nice - previous.Nice) / factor
	diff.System = (s.System - previous.System) / factor
	diff.Idle = (s.Idle - previous.Idle) / factor
	diff.IoWait = (s.IoWait - previous.IoWait) / factor
	diff.Irq = (s.Irq - previous.Irq) / factor
	diff.SoftIrq = (s.SoftIrq - previous.SoftIrq) / factor
	diff.Steal = (s.Steal - previous.Steal) / factor
	diff.Guest = (s.Guest - previous.Guest) / factor
	diff.GuestNice = (s.GuestNice - previous.GuestNice) / factor

	return &diff
}

func (c *CpuStats) GetMap(m map[string]interface{}) {
	m["misc.Interrupts"] = c.Interrupts
	m["misc.ContextSwitches"] = c.ContextSwitches
	m["misc.Forks"] = c.Forks
	m["misc.RunningProcesses"] = c.RunningProcesses
	m["misc.BlockedProcesses"] = c.BlockedProcesses

	for key, value := range c.Cpu {
		m[key+".User"] = value.User
		m[key+".Nice"] = value.Nice
		m[key+".System"] = value.System
		m[key+".Idle"] = value.Idle
		m[key+".IoWait"] = value.IoWait
		m[key+".Irq"] = value.Irq
		m[key+".SoftIrq"] = value.SoftIrq
		m[key+".Steal"] = value.Steal
		m[key+".Guest"] = value.Guest
		m[key+".GuestNice"] = value.GuestNice
	}
}

func (c *CpuStats) GetDoc(m map[string]string) {
	m["misc.Interrupts"] = "Number of interrupts per second (/s)"
	m["misc.ContextSwitches"] = "Number of context switches per second (/s)"
	m["misc.Forks"] = "Number of forks per second (/s)"
	m["misc.RunningProcesses"] = "Currently running processes (n)"
	m["misc.BlockedProcesses"] = "Number of processes currently blocked (n)"

	m["cpu.<n>.User"] = "Time spend in user mode (ticks/s)"
	m["cpu.<n>.Nice"] = "Time spend in user mode with low priority (ticks/s)"
	m["cpu.<n>.System"] = "Time spend in kernel mode (ticks/s)"
	m["cpu.<n>.Idle"] = "Time spend idle (ticks/s)"
	m["cpu.<n>.IoWait"] = "Time spend waiting for IO (ticks/s)"
	m["cpu.<n>.Irq"] = "Time spend processing interrupts (ticks/s)"
	m["cpu.<n>.SoftIrq"] = "Time spend processing soft interrupts (ticks/s)"
	m["cpu.<n>.Steal"] = "Time spend waiting for the *physical* CPU on a guest (ticks/s)"
	m["cpu.<n>.Guest"] = "Time spend on running guests (ticks/s)"
	m["cpu.<n>.GuestNice"] = "Time spend on running nice guests (ticks/s)"
}

type CpuSpeed struct {
	Frequency []int
}

func GetCpuSpeed() *CpuSpeed {
	speed := CpuSpeed{}

	files, err := filepath.Glob("/sys/devices/system/cpu/cpu[0-9]*/cpufreq/cpuinfo_max_freq")

	if err != nil {
		return &speed
	}

	speed.Frequency = make([]int, len(files))
	i := 0
	for _, file := range files {
		contents, err := ioutil.ReadFile(file)
		if err == nil {
			speed.Frequency[i], _ = strconv.Atoi(strings.TrimSpace(string(contents)))
			i += 1
		}
	}

	return &speed
}
func (c CpuSpeed) MarshalJSON() ([]byte, error) {
	return json.Marshal(c.Frequency)
}

func (c *CpuSpeed) UnmarshalJSON(b []byte) error {
	c.Frequency = make([]int, 100)
	return json.Unmarshal(b, &c.Frequency)
}

func (c *CpuSpeed) GetMap(m map[string]interface{}) {

	for i, frequency := range c.Frequency {
		m["cpu."+strconv.Itoa(i)+".Frequency"] = frequency
	}
}

func (c *CpuSpeed) GetDoc(m map[string]string) {

	m["cpu.<n>.Frequency"] = "The current CPU frequency (kHz)"
}
