package linux

import (
	"bufio"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"time"

	"github.com/abrander/agento"
)

var previousDiskStats *DiskStats

type DiskStats struct {
	sampletime time.Time                   `json:"-"`
	Disks      map[string]*SingleDiskStats `json:"disks"`
}

func GetDiskStats() *DiskStats {
	stat := DiskStats{}
	stat.Disks = make(map[string]*SingleDiskStats)

	path := filepath.Join("/proc/diskstats")
	file, err := os.Open(path)
	if err != nil {
		return &stat
	}
	defer file.Close()

	stat.sampletime = time.Now()
	scanner := bufio.NewScanner(file)
	for scanner.Scan() {
		text := scanner.Text()

		data := strings.Fields(strings.Trim(text, " "))
		if len(data) != 14 {
			continue
		}

		readsCompleted, _ := strconv.ParseInt(data[3], 10, 32)

		if readsCompleted > 0 {
			s := SingleDiskStats{}
			s.ReadArray(data)
			stat.Disks[data[2]] = &s
		}
	}

	ret := stat.Sub(previousDiskStats)
	previousDiskStats = &stat

	return ret
}

func (c *DiskStats) Sub(previousDiskStats *DiskStats) *DiskStats {
	if previousDiskStats == nil {
		return &DiskStats{}
	}

	diff := DiskStats{}
	diff.Disks = make(map[string]*SingleDiskStats)

	duration := float64(c.sampletime.Sub(previousDiskStats.sampletime)) / float64(time.Second)
	for key, value := range c.Disks {
		diff.Disks[key] = value.Sub(previousDiskStats.Disks[key], duration)
	}

	return &diff
}

func (c *DiskStats) GetMap(m map[string]interface{}) {
	if c == nil {
		return
	}

	if c.Disks == nil {
		return
	}

	for key, value := range c.Disks {
		m["io."+key+".ReadsCompleted"] = value.ReadsCompleted
		m["io."+key+".ReadsMerged"] = value.ReadsMerged
		m["io."+key+".ReadSectors"] = value.ReadSectors
		m["io."+key+".ReadTime"] = value.ReadTime
		m["io."+key+".WritesCompleted"] = value.WritesCompleted
		m["io."+key+".WritesMerged"] = value.WritesMerged
		m["io."+key+".WriteSectors"] = value.WriteSectors
		m["io."+key+".WriteTime"] = value.WriteTime
		m["io."+key+".IoInProgress"] = value.IoInProgress
		m["io."+key+".IoTime"] = value.IoTime
		m["io."+key+".IoWeightedTime"] = value.IoWeightedTime
	}
}

func (c *DiskStats) GetDoc(m map[string]string) {
	m["io.<device>.ReadsCompleted"] = "Reads from device (reads/s)"
	m["io.<device>.ReadsMerged"] = "Reads merged (merges/s)"
	m["io.<device>.ReadSectors"] = "Sectors read (sectors/s)"
	m["io.<device>.ReadTime"] = "Milliseconds spend reading (ms/s)"
	m["io.<device>.WritesCompleted"] = "Writes to device (writes/s)"
	m["io.<device>.WritesMerged"] = "Writes merged (merges/s)"
	m["io.<device>.WriteSectors"] = "Sectors written (sectors/s"
	m["io.<device>.WriteTime"] = "Time spend writing (ms/s)"
	m["io.<device>.IoInProgress"] = "The current queue size of IO operations (n)"
	m["io.<device>.IoTime"] = "Time spend on IO (ms/s)"
	m["io.<device>.IoWeightedTime"] = "Time spend on IO times the IO queue. Please see https://www.kernel.org/doc/Documentation/iostats.txt (ms/s)"
}

type SingleDiskStats struct {
	ReadsCompleted  float64
	ReadsMerged     float64
	ReadSectors     float64
	ReadTime        float64
	WritesCompleted float64
	WritesMerged    float64
	WriteSectors    float64
	WriteTime       float64
	IoInProgress    int64
	IoTime          float64
	IoWeightedTime  float64
}

func (s *SingleDiskStats) ReadArray(data []string) {
	l := len(data)

	if l > 3 {
		s.ReadsCompleted, _ = strconv.ParseFloat(data[3], 64)
	}

	if l > 4 {
		s.ReadsMerged, _ = strconv.ParseFloat(data[4], 64)
	}

	if l > 5 {
		s.ReadSectors, _ = strconv.ParseFloat(data[5], 64)
	}

	if l > 6 {
		s.ReadTime, _ = strconv.ParseFloat(data[6], 64)
	}

	if l > 7 {
		s.WritesCompleted, _ = strconv.ParseFloat(data[7], 64)
	}

	if l > 8 {
		s.WritesMerged, _ = strconv.ParseFloat(data[8], 64)
	}

	if l > 9 {
		s.WriteSectors, _ = strconv.ParseFloat(data[9], 64)
	}

	if l > 10 {
		s.WriteTime, _ = strconv.ParseFloat(data[10], 64)
	}

	if l > 11 {
		s.IoInProgress, _ = strconv.ParseInt(data[11], 10, 32)
	}

	if l > 12 {
		s.IoTime, _ = strconv.ParseFloat(data[12], 64)
	}

	if l > 13 {
		s.IoWeightedTime, _ = strconv.ParseFloat(data[13], 64)
	}
}

func (s SingleDiskStats) MarshalJSON() ([]byte, error) {
	var a [11]float64

	a[0] = Round(s.ReadsCompleted, 1)
	a[1] = Round(s.ReadsMerged, 1)
	a[2] = Round(s.ReadSectors, 1)
	a[3] = Round(s.ReadTime, 1)
	a[4] = Round(s.WritesCompleted, 1)
	a[5] = Round(s.WritesMerged, 1)
	a[6] = Round(s.WriteSectors, 1)
	a[7] = Round(s.WriteTime, 1)
	a[8] = Round(float64(s.IoInProgress), 0)
	a[9] = Round(s.IoTime, 1)
	a[10] = Round(s.IoWeightedTime, 1)

	return json.Marshal(a)
}

func (s *SingleDiskStats) UnmarshalJSON(b []byte) error {
	var a [11]float64

	err := json.Unmarshal(b, &a)
	if err != nil {
		return err
	}

	s.ReadsCompleted = a[0]
	s.ReadsMerged = a[1]
	s.ReadSectors = a[2]
	s.ReadTime = a[3]
	s.WritesCompleted = a[4]
	s.WritesMerged = a[5]
	s.WriteSectors = a[6]
	s.WriteTime = a[7]
	s.IoInProgress = int64(a[8])
	s.IoTime = a[9]
	s.IoWeightedTime = a[10]

	return err
}

func (s *SingleDiskStats) Sub(previous *SingleDiskStats, factor float64) *SingleDiskStats {
	diff := SingleDiskStats{}
	diff.IoInProgress = s.IoInProgress

	diff.ReadsCompleted = (s.ReadsCompleted - previous.ReadsCompleted) / factor
	diff.ReadsMerged = (s.ReadsMerged - previous.ReadsMerged) / factor
	diff.ReadSectors = (s.ReadSectors - previous.ReadSectors) / factor
	diff.ReadTime = (s.ReadTime - previous.ReadTime) / factor
	diff.WritesCompleted = (s.WritesCompleted - previous.WritesCompleted) / factor
	diff.WritesMerged = (s.WritesMerged - previous.WritesMerged) / factor
	diff.WriteSectors = (s.WriteSectors - previous.WriteSectors) / factor
	diff.WriteTime = (s.WriteTime - previous.WriteTime) / factor
	diff.IoTime = (s.IoTime - previous.IoTime) / factor
	diff.IoWeightedTime = (s.IoWeightedTime - previous.IoWeightedTime) / factor

	return &diff
}

type SingleDiskUsageStats struct {
	Used      uint64 `json:"u"`
	Reserved  uint64 `json:"r"`
	Free      uint64 `json:"f"`
	UsedNodes uint64 `json:"un"`
	FreeNodes uint64 `json:"fn"`
}

func ReadSingleDiskUsageStats(path string) *SingleDiskUsageStats {
	var stats SingleDiskUsageStats
	var stat syscall.Statfs_t

	err := syscall.Statfs(path, &stat)
	if err != nil {
		return nil
	}

	bSize := uint64(stat.Bsize)
	stats.Used = bSize * (stat.Blocks - stat.Bfree)
	stats.Reserved = bSize * (stat.Bfree - stat.Bavail)
	stats.Free = bSize * stat.Bavail

	stats.UsedNodes = stat.Files - stat.Ffree
	stats.FreeNodes = stat.Ffree

	return &stats
}

type DiskUsageStats struct {
	Disks map[string]*SingleDiskUsageStats `json:"disks"`
}

func GetDiskUsageStats() *DiskUsageStats {
	stat := DiskUsageStats{}
	stat.Disks = make(map[string]*SingleDiskUsageStats)

	stat.Disks["/"] = ReadSingleDiskUsageStats("/")

	return &stat
}

func (c *DiskUsageStats) GetMap(m map[string]interface{}) {
	if c == nil {
		return
	}

	if c.Disks == nil {
		return
	}

	for key, value := range c.Disks {
		m["du."+key+".Used"] = value.Used
		m["du."+key+".Reserved"] = value.Reserved
		m["du."+key+".Free"] = value.Free
		m["du."+key+".UsedNodes"] = value.UsedNodes
		m["du."+key+".FreeNodes"] = value.FreeNodes
	}
}

func (c *DiskUsageStats) GetDoc(m map[string]string) {
	m["du.<mountpoint>.Used"] = "Used space (b)"
	m["du.<mountpoint>.Reserved"] = "Space reserved for uid 0 (b)"
	m["du.<mountpoint>.Free"] = "Free space (b)"
	m["du.<mountpoint>.UsedNodes"] = "Used inodes (n)"
	m["du.<mountpoint>.FreeNodes"] = "Free inodes (n)"
}
