// Copyright 2010 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package math32

/*
	Floating-point error function and complementary error function.
*/

// The original C code and the long comment below are
// from FreeBSD's /usr/src/lib/msun/src/s_erf.c and
// came with this notice.  The go code is a simplified
// version of the original C.
//
// ====================================================
// Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
//
// Developed at SunPro, a Sun Microsystems, Inc. business.
// Permission to use, copy, modify, and distribute this
// software is freely granted, provided that this notice
// is preserved.
// ====================================================
//
//
// double erf(double x)
// double erfc(double x)
//                           x
//                    2      |\
//     erf(x)  =  ---------  | exp(-t*t)dt
//                 sqrt(pi) \|
//                           0
//
//     erfc(x) =  1-erf(x)
//  Note that
//              erf(-x) = -erf(x)
//              erfc(-x) = 2 - erfc(x)
//
// Method:
//      1. For |x| in [0, 0.84375]
//          erf(x)  = x + x*R(x**2)
//          erfc(x) = 1 - erf(x)           if x in [-.84375,0.25]
//                  = 0.5 + ((0.5-x)-x*R)  if x in [0.25,0.84375]
//         where R = P/Q where P is an odd poly of degree 8 and
//         Q is an odd poly of degree 10.
//                                               -57.90
//                      | R - (erf(x)-x)/x | <= 2
//
//
//         Remark. The formula is derived by noting
//          erf(x) = (2/sqrt(pi))*(x - x**3/3 + x**5/10 - x**7/42 + ....)
//         and that
//          2/sqrt(pi) = 1.128379167095512573896158903121545171688
//         is close to one. The interval is chosen because the fix
//         point of erf(x) is near 0.6174 (i.e., erf(x)=x when x is
//         near 0.6174), and by some experiment, 0.84375 is chosen to
//         guarantee the error is less than one ulp for erf.
//
//      2. For |x| in [0.84375,1.25], let s = |x| - 1, and
//         c = 0.84506291151 rounded to single (24 bits)
//              erf(x)  = sign(x) * (c  + P1(s)/Q1(s))
//              erfc(x) = (1-c)  - P1(s)/Q1(s) if x > 0
//                        1+(c+P1(s)/Q1(s))    if x < 0
//              |P1/Q1 - (erf(|x|)-c)| <= 2**-59.06
//         Remark: here we use the taylor series expansion at x=1.
//              erf(1+s) = erf(1) + s*Poly(s)
//                       = 0.845.. + P1(s)/Q1(s)
//         That is, we use rational approximation to approximate
//                      erf(1+s) - (c = (single)0.84506291151)
//         Note that |P1/Q1|< 0.078 for x in [0.84375,1.25]
//         where
//              P1(s) = degree 6 poly in s
//              Q1(s) = degree 6 poly in s
//
//      3. For x in [1.25,1/0.35(~2.857143)],
//              erfc(x) = (1/x)*exp(-x*x-0.5625+R1/S1)
//              erf(x)  = 1 - erfc(x)
//         where
//              R1(z) = degree 7 poly in z, (z=1/x**2)
//              S1(z) = degree 8 poly in z
//
//      4. For x in [1/0.35,28]
//              erfc(x) = (1/x)*exp(-x*x-0.5625+R2/S2) if x > 0
//                      = 2.0 - (1/x)*exp(-x*x-0.5625+R2/S2) if -6<x<0
//                      = 2.0 - tiny            (if x <= -6)
//              erf(x)  = sign(x)*(1.0 - erfc(x)) if x < 6, else
//              erf(x)  = sign(x)*(1.0 - tiny)
//         where
//              R2(z) = degree 6 poly in z, (z=1/x**2)
//              S2(z) = degree 7 poly in z
//
//      Note1:
//         To compute exp(-x*x-0.5625+R/S), let s be a single
//         precision number and s := x; then
//              -x*x = -s*s + (s-x)*(s+x)
//              exp(-x*x-0.5626+R/S) =
//                      exp(-s*s-0.5625)*exp((s-x)*(s+x)+R/S);
//      Note2:
//         Here 4 and 5 make use of the asymptotic series
//                        exp(-x*x)
//              erfc(x) ~ ---------- * ( 1 + Poly(1/x**2) )
//                        x*sqrt(pi)
//         We use rational approximation to approximate
//              g(s)=f(1/x**2) = log(erfc(x)*x) - x*x + 0.5625
//         Here is the error bound for R1/S1 and R2/S2
//              |R1/S1 - f(x)|  < 2**(-62.57)
//              |R2/S2 - f(x)|  < 2**(-61.52)
//
//      5. For inf > x >= 28
//              erf(x)  = sign(x) *(1 - tiny)  (raise inexact)
//              erfc(x) = tiny*tiny (raise underflow) if x > 0
//                      = 2 - tiny if x<0
//
//      7. Special case:
//              erf(0)  = 0, erf(inf)  = 1, erf(-inf) = -1,
//              erfc(0) = 1, erfc(inf) = 0, erfc(-inf) = 2,
//              erfc/erf(NaN) is NaN

const (
	erx = 8.4506291151e-01 /* 0x3f58560b */
	// Coefficients for approximation to  erf in [0, 0.84375]
	efx  = 1.2837916613e-01  /* 0x3e0375d4 */
	efx8 = 1.0270333290e+00  /* 0x3f8375d4 */
	pp0  = 1.2837916613e-01  /* 0x3e0375d4 */
	pp1  = -3.2504209876e-01 /* 0xbea66beb */
	pp2  = -2.8481749818e-02 /* 0xbce9528f */
	pp3  = -5.7702702470e-03 /* 0xbbbd1489 */
	pp4  = -2.3763017452e-05 /* 0xb7c756b1 */
	qq1  = 3.9791721106e-01  /* 0x3ecbbbce */
	qq2  = 6.5022252500e-02  /* 0x3d852a63 */
	qq3  = 5.0813062117e-03  /* 0x3ba68116 */
	qq4  = 1.3249473704e-04  /* 0x390aee49 */
	qq5  = -3.9602282413e-06 /* 0xb684e21a */
	// Coefficients for approximation to  erf  in [0.84375, 1.25]
	pa0 = -2.3621185683e-03 /* 0xbb1acdc6 */
	pa1 = 4.1485610604e-01  /* 0x3ed46805 */
	pa2 = -3.7220788002e-01 /* 0xbebe9208 */
	pa3 = 3.1834661961e-01  /* 0x3ea2fe54 */
	pa4 = -1.1089469492e-01 /* 0xbde31cc2 */
	pa5 = 3.5478305072e-02  /* 0x3d1151b3 */
	pa6 = -2.1663755178e-03 /* 0xbb0df9c0 */
	qa1 = 1.0642088205e-01  /* 0x3dd9f331 */
	qa2 = 5.4039794207e-01  /* 0x3f0a5785 */
	qa3 = 7.1828655899e-02  /* 0x3d931ae7 */
	qa4 = 1.2617121637e-01  /* 0x3e013307 */
	qa5 = 1.3637083583e-02  /* 0x3c5f6e13 */
	qa6 = 1.1984500103e-02  /* 0x3c445aa3 */
	// Coefficients for approximation to  erfc in [1.25, 1/0.35]
	ra0 = -9.8649440333e-03 /* 0xbc21a093 */
	ra1 = -6.9385856390e-01 /* 0xbf31a0b7 */
	ra2 = -1.0558626175e+01 /* 0xc128f022 */
	ra3 = -6.2375331879e+01 /* 0xc2798057 */
	ra4 = -1.6239666748e+02 /* 0xc322658c */
	ra5 = -1.8460508728e+02 /* 0xc3389ae7 */
	ra6 = -8.1287437439e+01 /* 0xc2a2932b */
	ra7 = -9.8143291473e+00 /* 0xc11d077e */
	sa1 = 1.9651271820e+01  /* 0x419d35ce */
	sa2 = 1.3765776062e+02  /* 0x4309a863 */
	sa3 = 4.3456588745e+02  /* 0x43d9486f */
	sa4 = 6.4538726807e+02  /* 0x442158c9 */
	sa5 = 4.2900814819e+02  /* 0x43d6810b */
	sa6 = 1.0863500214e+02  /* 0x42d9451f */
	sa7 = 6.5702495575e+00  /* 0x40d23f7c */
	sa8 = -6.0424413532e-02 /* 0xbd777f97 */
	// Coefficients for approximation to  erfc in [1/.35, 28]
	rb0 = -9.8649431020e-03 /* 0xbc21a092 */
	rb1 = -7.9928326607e-01 /* 0xbf4c9dd4 */
	rb2 = -1.7757955551e+01 /* 0xc18e104b */
	rb3 = -1.6063638306e+02 /* 0xc320a2ea */
	rb4 = -6.3756646729e+02 /* 0xc41f6441 */
	rb5 = -1.0250950928e+03 /* 0xc480230b */
	rb6 = -4.8351919556e+02 /* 0xc3f1c275 */
	sb1 = 3.0338060379e+01  /* 0x41f2b459 */
	sb2 = 3.2579251099e+02  /* 0x43a2e571 */
	sb3 = 1.5367296143e+03  /* 0x44c01759 */
	sb4 = 3.1998581543e+03  /* 0x4547fdbb */
	sb5 = 2.5530502930e+03  /* 0x451f90ce */
	sb6 = 4.7452853394e+02  /* 0x43ed43a7 */
	sb7 = -2.2440952301e+01 /* 0xc1b38712 */
)

// Erf returns the error function of x.
//
// Special cases are:
//	Erf(+Inf) = 1
//	Erf(-Inf) = -1
//	Erf(NaN) = NaN
func Erf(x float32) float32 {
	const (
		// VeryTiny = 2.848094538889218e-306 // 0x0080000000000000
		VeryTiny = 2.3509887e-38   // 0x01000000
		Small    = 1.0 / (1 << 28) // 2**-28
	)
	// special cases
	switch {
	case IsNaN(x):
		return NaN()
	case IsInf(x, 1):
		return 1
	case IsInf(x, -1):
		return -1
	}
	sign := false
	if x < 0 {
		x = -x
		sign = true
	}
	if x < 0.84375 { // |x| < 0.84375
		var temp float32
		if x < Small { // |x| < 2**-28
			if x < VeryTiny {
				temp = 0.125 * (8.0*x + efx8*x) // avoid underflow
			} else {
				temp = x + efx*x
			}
		} else {
			z := x * x
			r := pp0 + z*(pp1+z*(pp2+z*(pp3+z*pp4)))
			s := 1 + z*(qq1+z*(qq2+z*(qq3+z*(qq4+z*qq5))))
			y := r / s
			temp = x + x*y
		}
		if sign {
			return -temp
		}
		return temp
	}
	if x < 1.25 { // 0.84375 <= |x| < 1.25
		s := x - 1
		P := pa0 + s*(pa1+s*(pa2+s*(pa3+s*(pa4+s*(pa5+s*pa6)))))
		Q := 1 + s*(qa1+s*(qa2+s*(qa3+s*(qa4+s*(qa5+s*qa6)))))
		if sign {
			return -erx - P/Q
		}
		return erx + P/Q
	}
	if x >= 6 { // inf > |x| >= 6
		if sign {
			return -1
		}
		return 1
	}
	s := 1 / (x * x)
	var R, S float32
	if x < 1/0.35 { // |x| < 1 / 0.35  ~ 2.857143
		R = ra0 + s*(ra1+s*(ra2+s*(ra3+s*(ra4+s*(ra5+s*(ra6+s*ra7))))))
		S = 1 + s*(sa1+s*(sa2+s*(sa3+s*(sa4+s*(sa5+s*(sa6+s*(sa7+s*sa8)))))))
	} else { // |x| >= 1 / 0.35  ~ 2.857143
		R = rb0 + s*(rb1+s*(rb2+s*(rb3+s*(rb4+s*(rb5+s*rb6)))))
		S = 1 + s*(sb1+s*(sb2+s*(sb3+s*(sb4+s*(sb5+s*(sb6+s*sb7))))))
	}
	z := Float32frombits(Float32bits(x) & 0xfffff000)
	r := Exp(-z*z-0.5625) * Exp((z-x)*(z+x)+R/S)
	if sign {
		return r/x - 1
	}
	return 1 - r/x
}

// Erfc returns the complementary error function of x.
//
// Special cases are:
//	Erfc(+Inf) = 0
//	Erfc(-Inf) = 2
//	Erfc(NaN) = NaN
func Erfc(x float32) float32 {
	const Tiny = 1.0 / (1 << 56) // 2**-56
	// special cases
	switch {
	case IsNaN(x):
		return NaN()
	case IsInf(x, 1):
		return 0
	case IsInf(x, -1):
		return 2
	}
	sign := false
	if x < 0 {
		x = -x
		sign = true
	}
	if x < 0.84375 { // |x| < 0.84375
		var temp float32
		if x < Tiny { // |x| < 2**-56
			temp = x
		} else {
			z := x * x
			r := pp0 + z*(pp1+z*(pp2+z*(pp3+z*pp4)))
			s := 1 + z*(qq1+z*(qq2+z*(qq3+z*(qq4+z*qq5))))
			y := r / s
			if x < 0.25 { // |x| < 1/4
				temp = x + x*y
			} else {
				temp = 0.5 + (x*y + (x - 0.5))
			}
		}
		if sign {
			return 1 + temp
		}
		return 1 - temp
	}
	if x < 1.25 { // 0.84375 <= |x| < 1.25
		s := x - 1
		P := pa0 + s*(pa1+s*(pa2+s*(pa3+s*(pa4+s*(pa5+s*pa6)))))
		Q := 1 + s*(qa1+s*(qa2+s*(qa3+s*(qa4+s*(qa5+s*qa6)))))
		if sign {
			return 1 + erx + P/Q
		}
		return 1 - erx - P/Q

	}
	if x < 28 { // |x| < 28
		s := 1 / (x * x)
		var R, S float32
		if x < 1/0.35 { // |x| < 1 / 0.35 ~ 2.857143
			R = ra0 + s*(ra1+s*(ra2+s*(ra3+s*(ra4+s*(ra5+s*(ra6+s*ra7))))))
			S = 1 + s*(sa1+s*(sa2+s*(sa3+s*(sa4+s*(sa5+s*(sa6+s*(sa7+s*sa8)))))))
		} else { // |x| >= 1 / 0.35 ~ 2.857143
			if sign && x > 6 {
				return 2 // x < -6
			}
			R = rb0 + s*(rb1+s*(rb2+s*(rb3+s*(rb4+s*(rb5+s*rb6)))))
			S = 1 + s*(sb1+s*(sb2+s*(sb3+s*(sb4+s*(sb5+s*(sb6+s*sb7))))))
		}
		z := Float32frombits(Float32bits(x) & 0xfffff000)
		r := Exp(-z*z-0.5625) * Exp((z-x)*(z+x)+R/S)
		if sign {
			return 2 - r/x
		}
		return r / x
	}
	if sign {
		return 2
	}
	return 0
}
