package agento

import (
	"os"
	"sync"
	"time"

	"github.com/abrander/agento/plugins"
)

type Plugin interface {
	Gather() Plugin
	Map(map[string]interface{})
	Doc(map[string]interface{})
}

type MachineStats struct {
	Hostname         string                  `json:"h"`
	MemoryStats      *plugins.MemoryStats    `json:"m"`
	CpuStats         *plugins.CpuStats       `json:"c"`
	DiskStats        *plugins.DiskStats      `json:"d"`
	DiskUsageStats   *plugins.DiskUsageStats `json:"du"`
	NetStats         *plugins.NetStats       `json:"n"`
	LoadStats        *plugins.LoadStats      `json:"l"`
	SnmpStats        *plugins.SnmpStats      `json:"s"`
	CpuSpeed         *plugins.CpuSpeed       `json:"f"`
	AvailableEntropy int64                   `json:"e"`
	GatherDuration   time.Duration           `json:"g"`
	plugins          []Plugin
}

type Gatherer struct {
}

func NewMachineStats() *MachineStats {
	var m = MachineStats{}
	m.Gather()
	return &m
}

func (m *MachineStats) Gather() {
	start := time.Now()
	m.Hostname, _ = os.Hostname()
	m.MemoryStats = plugins.MemoryStats()
	m.CpuStats = plugins.CpuStats()
	m.DiskStats = plugins.DiskStats()
	m.DiskUsageStats = plugins.DiskUsageStats()
	m.NetStats = plugins.NetStats()
	m.LoadStats = plugins.LoadStats()
	m.SnmpStats = plugins.SnmpStats()
	m.CpuSpeed = plugins.CpuSpeed()
	m.AvailableEntropy, _ = getAvailableEntropy()
	m.GatherDuration = time.Now().Sub(start)
}

func (s *MachineStats) Map() map[string]interface{} {
	m := make(map[string]interface{})
	s.MemoryStats.GetMap(m)
	s.CpuStats.GetMap(m)
	s.DiskStats.GetMap(m)
	s.DiskUsageStats.GetMap(m)
	s.NetStats.GetMap(m)
	s.LoadStats.GetMap(m)
	s.SnmpStats.GetMap(m)
	s.CpuSpeed.GetMap(m)
	m["misc.AvailableEntropy"] = s.AvailableEntropy
	m["agento.GatherDuration"] = round(s.GatherDuration.Seconds()*1000.0, 1)
	return m
}

func (s *MachineStats) Doc() map[string]string {
	m := make(map[string]string)
	s.MemoryStats.GetDoc(m)
	s.CpuStats.GetDoc(m)
	s.DiskStats.GetDoc(m)
	s.DiskUsageStats.GetDoc(m)
	s.NetStats.GetDoc(m)
	s.LoadStats.GetDoc(m)
	s.SnmpStats.GetDoc(m)
	s.CpuSpeed.GetDoc(m)
	m["misc.AvailableEntropy"] = "Available entropy in the kernel pool (b)"
	m["agento.GatherDuration"] = "Time used to gather metrics for Agento (ms)"
	return m
}

func getAvailableEntropy() (int64, error) {
	contents, err := ioutil.ReadFile("/proc/sys/kernel/random/entropy_avail")
	if err != nil {
		return 0, err
	}
	return strconv.ParseInt(strings.TrimSpace(string(contents)), 10, 64)
}

func round(value float64, places int) float64 {
	var round float64
	pow := math.Pow(10, float64(places))
	digit := pow * value
	_, div := math.Modf(digit)
	if div >= 0.5 {
		round = math.Ceil(digit)
	} else {
		round = math.Floor(digit)
	}
	return round / pow
}
