// Copyright 2010 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package math32

/*
	Bessel function of the first and second kinds of order zero.
*/

// The original C code and the long comment below are
// from FreeBSD's /usr/src/lib/msun/src/e_j0.c and
// came with this notice.  The go code is a simplified
// version of the original C.
//
// ====================================================
// Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
//
// Developed at SunPro, a Sun Microsystems, Inc. business.
// Permission to use, copy, modify, and distribute this
// software is freely granted, provided that this notice
// is preserved.
// ====================================================
//
// __ieee754_j0(x), __ieee754_y0(x)
// Bessel function of the first and second kinds of order zero.
// Method -- j0(x):
//      1. For tiny x, we use j0(x) = 1 - x**2/4 + x**4/64 - ...
//      2. Reduce x to |x| since j0(x)=j0(-x),  and
//         for x in (0,2)
//              j0(x) = 1-z/4+ z**2*R0/S0,  where z = x*x;
//         (precision:  |j0-1+z/4-z**2R0/S0 |<2**-63.67 )
//         for x in (2,inf)
//              j0(x) = sqrt(2/(pi*x))*(p0(x)*cos(x0)-q0(x)*sin(x0))
//         where x0 = x-pi/4. It is better to compute sin(x0),cos(x0)
//         as follow:
//              cos(x0) = cos(x)cos(pi/4)+sin(x)sin(pi/4)
//                      = 1/sqrt(2) * (cos(x) + sin(x))
//              sin(x0) = sin(x)cos(pi/4)-cos(x)sin(pi/4)
//                      = 1/sqrt(2) * (sin(x) - cos(x))
//         (To avoid cancellation, use
//              sin(x) +- cos(x) = -cos(2x)/(sin(x) -+ cos(x))
//         to compute the worse one.)
//
//      3 Special cases
//              j0(nan)= nan
//              j0(0) = 1
//              j0(inf) = 0
//
// Method -- y0(x):
//      1. For x<2.
//         Since
//              y0(x) = 2/pi*(j0(x)*(ln(x/2)+Euler) + x**2/4 - ...)
//         therefore y0(x)-2/pi*j0(x)*ln(x) is an even function.
//         We use the following function to approximate y0,
//              y0(x) = U(z)/V(z) + (2/pi)*(j0(x)*ln(x)), z= x**2
//         where
//              U(z) = u00 + u01*z + ... + u06*z**6
//              V(z) = 1  + v01*z + ... + v04*z**4
//         with absolute approximation error bounded by 2**-72.
//         Note: For tiny x, U/V = u0 and j0(x)~1, hence
//              y0(tiny) = u0 + (2/pi)*ln(tiny), (choose tiny<2**-27)
//      2. For x>=2.
//              y0(x) = sqrt(2/(pi*x))*(p0(x)*cos(x0)+q0(x)*sin(x0))
//         where x0 = x-pi/4. It is better to compute sin(x0),cos(x0)
//         by the method mentioned above.
//      3. Special cases: y0(0)=-inf, y0(x<0)=NaN, y0(inf)=0.
//

// J0 returns the order-zero Bessel function of the first kind.
//
// Special cases are:
//	J0(±Inf) = 0
//	J0(0) = 1
//	J0(NaN) = NaN
func J0(x float32) float32 {
	const (
		Huge   = 1e30
		TwoM27 = 1.0 / (1 << 27) // 2**-27 0x3e40000000000000
		TwoM13 = 1.0 / (1 << 13) // 2**-13 0x3f20000000000000
		// R0/S0 on [0, 2]
		R02 = 1.5625000000e-02  // 0x3c800000
		R03 = -1.8997929874e-04 // 0xb947352e
		R04 = 1.8295404516e-06  // 0x35f58e88
		R05 = -4.6183270541e-09 // 0xb19eaf3c
		S01 = 1.5619102865e-02  // 0x3c7fe744
		S02 = 1.1692678527e-04  // 0x38f53697
		S03 = 5.1354652442e-07  // 0x3509daa6
		S04 = 1.1661400734e-09  // 0x30a045e8

	)
	// special cases
	switch {
	case IsNaN(x):
		return x
	case IsInf(x, 0):
		return 0
	case x == 0:
		return 1
	}

	if x < 0 {
		x = -x
	}
	if x >= 2 {
		s, c := Sincos(x)
		ss := s - c
		cc := s + c

		// make sure x+x does not overflow
		if x < MaxFloat32/2 {
			z := -Cos(x + x)
			if s*c < 0 {
				cc = z / ss
			} else {
				ss = z / cc
			}
		}

		// j0(x) = 1/sqrt(pi) * (P(0,x)*cc - Q(0,x)*ss) / sqrt(x)
		// y0(x) = 1/sqrt(pi) * (P(0,x)*ss + Q(0,x)*cc) / sqrt(x)

		var z float32
		u := pzero(x)
		v := qzero(x)
		z = (1 / SqrtPi) * (u*cc - v*ss) / Sqrt(x)
		return z // |x| >= 2.0
	}
	if x < TwoM13 { // |x| < ~1.2207e-4
		if x < TwoM27 {
			return 1 // |x| < ~7.4506e-9
		}
		return 1 - 0.25*x*x // ~7.4506e-9 < |x| < ~1.2207e-4
	}
	z := x * x
	r := z * (R02 + z*(R03+z*(R04+z*R05)))
	s := 1 + z*(S01+z*(S02+z*(S03+z*S04)))
	if x < 1 {
		return 1 + z*(-0.25+(r/s)) // |x| < 1.00
	}
	u := 0.5 * x
	return (1+u)*(1-u) + z*(r/s) // 1.0 < |x| < 2.0
}

// Y0 returns the order-zero Bessel function of the second kind.
//
// Special cases are:
//	Y0(+Inf) = 0
//	Y0(0) = -Inf
//	Y0(x < 0) = NaN
//	Y0(NaN) = NaN
func Y0(x float32) float32 {
	const (
		TwoM27 = 1.0 / (1 << 27)             // 2**-27 0x3e40000000000000
		Two129 = 1 << 129                    // 2**129 0x4800000000000000
		U00    = -7.38042951086872317523e-02 // 0xBFB2E4D699CBD01F
		U01    = 1.76666452509181115538e-01  // 0x3FC69D019DE9E3FC
		U02    = -1.38185671945596898896e-02 // 0xBF8C4CE8B16CFA97
		U03    = 3.47453432093683650238e-04  // 0x3F36C54D20B29B6B
		U04    = -3.81407053724364161125e-06 // 0xBECFFEA773D25CAD
		U05    = 1.95590137035022920206e-08  // 0x3E5500573B4EABD4
		U06    = -3.98205194132103398453e-11 // 0xBDC5E43D693FB3C8
		V01    = 1.27304834834123699328e-02  // 0x3F8A127091C9C71A
		V02    = 7.60068627350353253702e-05  // 0x3F13ECBBF578C6C1
		V03    = 2.59150851840457805467e-07  // 0x3E91642D7FF202FD
		V04    = 4.41110311332675467403e-10  // 0x3DFE50183BD6D9EF
	)
	// special cases
	switch {
	case x < 0 || IsNaN(x):
		return NaN()
	case IsInf(x, 1):
		return 0
	case x == 0:
		return Inf(-1)
	}

	if x >= 2 { // |x| >= 2.0

		// y0(x) = sqrt(2/(pi*x))*(p0(x)*sin(x0)+q0(x)*cos(x0))
		//     where x0 = x-pi/4
		// Better formula:
		//     cos(x0) = cos(x)cos(pi/4)+sin(x)sin(pi/4)
		//             =  1/sqrt(2) * (sin(x) + cos(x))
		//     sin(x0) = sin(x)cos(3pi/4)-cos(x)sin(3pi/4)
		//             =  1/sqrt(2) * (sin(x) - cos(x))
		// To avoid cancellation, use
		//     sin(x) +- cos(x) = -cos(2x)/(sin(x) -+ cos(x))
		// to compute the worse one.

		s, c := Sincos(x)
		ss := s - c
		cc := s + c

		// j0(x) = 1/sqrt(pi) * (P(0,x)*cc - Q(0,x)*ss) / sqrt(x)
		// y0(x) = 1/sqrt(pi) * (P(0,x)*ss + Q(0,x)*cc) / sqrt(x)

		// make sure x+x does not overflow
		if x < MaxFloat32/2 {
			z := -Cos(x + x)
			if s*c < 0 {
				cc = z / ss
			} else {
				ss = z / cc
			}
		}
		var z float32
		u := pzero(x)
		v := qzero(x)
		z = (1 / SqrtPi) * (u*ss + v*cc) / Sqrt(x)
		return z // |x| >= 2.0
	}
	if x <= TwoM27 {
		return U00 + (2/Pi)*Log(x) // |x| < ~7.4506e-9
	}
	z := x * x
	u := U00 + z*(U01+z*(U02+z*(U03+z*(U04+z*(U05+z*U06)))))
	v := 1 + z*(V01+z*(V02+z*(V03+z*V04)))
	return u/v + (2/Pi)*J0(x)*Log(x) // ~7.4506e-9 < |x| < 2.0
}

// The asymptotic expansions of pzero is
//      1 - 9/128 s**2 + 11025/98304 s**4 - ..., where s = 1/x.
// For x >= 2, We approximate pzero by
// 	pzero(x) = 1 + (R/S)
// where  R = pR0 + pR1*s**2 + pR2*s**4 + ... + pR5*s**10
// 	  S = 1 + pS0*s**2 + ... + pS4*s**10
// and
//      | pzero(x)-1-R/S | <= 2  ** ( -60.26)

// for x in [inf, 8]=1/[0,0.125]
var p0R8 = [6]float32{
	0.0000000000e+00,  /* 0x00000000 */
	-7.0312500000e-02, /* 0xbd900000 */
	-8.0816707611e+00, /* 0xc1014e86 */
	-2.5706311035e+02, /* 0xc3808814 */
	-2.4852163086e+03, /* 0xc51b5376 */
	-5.2530439453e+03, /* 0xc5a4285a */
}
var p0S8 = [5]float32{
	1.1653436279e+02, /* 0x42e91198 */
	3.8337448730e+03, /* 0x456f9beb */
	4.0597855469e+04, /* 0x471e95db */
	1.1675296875e+05, /* 0x47e4087c */
	4.7627726562e+04, /* 0x473a0bba */
}

// for x in [8,4.5454]=1/[0.125,0.22001]
var p0R5 = [6]float32{
	-1.1412546255e-11, /* 0xad48c58a */
	-7.0312492549e-02, /* 0xbd8fffff */
	-4.1596107483e+00, /* 0xc0851b88 */
	-6.7674766541e+01, /* 0xc287597b */
	-3.3123129272e+02, /* 0xc3a59d9b */
	-3.4643338013e+02, /* 0xc3ad3779 */
}
var p0S5 = [5]float32{
	6.0753936768e+01, /* 0x42730408 */
	1.0512523193e+03, /* 0x44836813 */
	5.9789707031e+03, /* 0x45bad7c4 */
	9.6254453125e+03, /* 0x461665c8 */
	2.4060581055e+03, /* 0x451660ee */
}

// for x in [4.547,2.8571]=1/[0.2199,0.35001]
var p0R3 = [6]float32{
	-2.5470459075e-09, /* 0xb12f081b */
	-7.0311963558e-02, /* 0xbd8fffb8 */
	-2.4090321064e+00, /* 0xc01a2d95 */
	-2.1965976715e+01, /* 0xc1afba52 */
	-5.8079170227e+01, /* 0xc2685112 */
	-3.1447946548e+01, /* 0xc1fb9565 */
}
var p0S3 = [5]float32{
	3.5856033325e+01, /* 0x420f6c94 */
	3.6151397705e+02, /* 0x43b4c1ca */
	1.1936077881e+03, /* 0x44953373 */
	1.1279968262e+03, /* 0x448cffe6 */
	1.7358093262e+02, /* 0x432d94b8 */
}

// for x in [2.8570,2]=1/[0.3499,0.5]
var p0R2 = [6]float32{
	-8.8753431271e-08, /* 0xb3be98b7 */
	-7.0303097367e-02, /* 0xbd8ffb12 */
	-1.4507384300e+00, /* 0xbfb9b1cc */
	-7.6356959343e+00, /* 0xc0f4579f */
	-1.1193166733e+01, /* 0xc1331736 */
	-3.2336456776e+00, /* 0xc04ef40d */
}
var p0S2 = [5]float32{
	2.2220300674e+01, /* 0x41b1c32d */
	1.3620678711e+02, /* 0x430834f0 */
	2.7047027588e+02, /* 0x43873c32 */
	1.5387539673e+02, /* 0x4319e01a */
	1.4657617569e+01, /* 0x416a859a */
}

func pzero(x float32) float32 {
	var p [6]float32
	var q [5]float32
	if x >= 8 {
		p = p0R8
		q = p0S8
	} else if x >= 7.7222 {
		p = p0R5
		q = p0S5
	} else if x >= 2.8571 {
		p = p0R3
		q = p0S3
	} else if x >= 2 {
		p = p0R2
		q = p0S2
	}
	z := 1 / (x * x)
	r := p[0] + z*(p[1]+z*(p[2]+z*(p[3]+z*(p[4]+z*p[5]))))
	s := 1 + z*(q[0]+z*(q[1]+z*(q[2]+z*(q[3]+z*q[4]))))
	return 1 + r/s
}

// For x >= 8, the asymptotic expansions of qzero is
//      -1/8 s + 75/1024 s**3 - ..., where s = 1/x.
// We approximate pzero by
//      qzero(x) = s*(-1.25 + (R/S))
// where R = qR0 + qR1*s**2 + qR2*s**4 + ... + qR5*s**10
//       S = 1 + qS0*s**2 + ... + qS5*s**12
// and
//      | qzero(x)/s +1.25-R/S | <= 2**(-61.22)

// for x in [inf, 8]=1/[0,0.125]
var q0R8 = [6]float32{
	0.0000000000e+00, /* 0x00000000 */
	7.3242187500e-02, /* 0x3d960000 */
	1.1768206596e+01, /* 0x413c4a93 */
	5.5767340088e+02, /* 0x440b6b19 */
	8.8591972656e+03, /* 0x460a6cca */
	3.7014625000e+04, /* 0x471096a0 */
}
var q0S8 = [6]float32{
	1.6377603149e+02,  /* 0x4323c6aa */
	8.0983447266e+03,  /* 0x45fd12c2 */
	1.4253829688e+05,  /* 0x480b3293 */
	8.0330925000e+05,  /* 0x49441ed4 */
	8.4050156250e+05,  /* 0x494d3359 */
	-3.4389928125e+05, /* 0xc8a7eb69 */
}

// for x in [8,4.5454]=1/[0.125,0.22001]
var q0R5 = [6]float32{
	1.8408595828e-11, /* 0x2da1ec79 */
	7.3242180049e-02, /* 0x3d95ffff */
	5.8356351852e+00, /* 0x40babd86 */
	1.3511157227e+02, /* 0x43071c90 */
	1.0272437744e+03, /* 0x448067cd */
	1.9899779053e+03, /* 0x44f8bf4b */
}
var q0S5 = [6]float32{
	8.2776611328e+01,  /* 0x42a58da0 */
	2.0778142090e+03,  /* 0x4501dd07 */
	1.8847289062e+04,  /* 0x46933e94 */
	5.6751113281e+04,  /* 0x475daf1d */
	3.5976753906e+04,  /* 0x470c88c1 */
	-5.3543427734e+03, /* 0xc5a752be */
}

// for x in [4.547,2.8571]=1/[0.2199,0.35001]
var q0R3 = [6]float32{
	4.3774099900e-09, /* 0x3196681b */
	7.3241114616e-02, /* 0x3d95ff70 */
	3.3442313671e+00, /* 0x405607e3 */
	4.2621845245e+01, /* 0x422a7cc5 */
	1.7080809021e+02, /* 0x432acedf */
	1.6673394775e+02, /* 0x4326bbe4 */
}
var q0S3 = [6]float32{
	4.8758872986e+01,  /* 0x42430916 */
	7.0968920898e+02,  /* 0x44316c1c */
	3.7041481934e+03,  /* 0x4567825f */
	6.4604252930e+03,  /* 0x45c9e367 */
	2.5163337402e+03,  /* 0x451d4557 */
	-1.4924745178e+02, /* 0xc3153f59 */
}

// for x in [2.8570,2]=1/[0.3499,0.5]
var q0R2 = [6]float32{
	1.5044444979e-07, /* 0x342189db */
	7.3223426938e-02, /* 0x3d95f62a */
	1.9981917143e+00, /* 0x3fffc4bf */
	1.4495602608e+01, /* 0x4167edfd */
	3.1666231155e+01, /* 0x41fd5471 */
	1.6252708435e+01, /* 0x4182058c */
}
var q0S2 = [6]float32{
	3.0365585327e+01,  /* 0x41f2ecb8 */
	2.6934811401e+02,  /* 0x4386ac8f */
	8.4478375244e+02,  /* 0x44533229 */
	8.8293585205e+02,  /* 0x445cbbe5 */
	2.1266638184e+02,  /* 0x4354aa98 */
	-5.3109550476e+00, /* 0xc0a9f358 */
}

func qzero(x float32) float32 {
	var p, q [6]float32
	if x >= 8 {
		p = q0R8
		q = q0S8
	} else if x >= 7.7222 {
		p = q0R5
		q = q0S5
	} else if x >= 2.8571 {
		p = q0R3
		q = q0S3
	} else if x >= 2 {
		p = q0R2
		q = q0S2
	}
	z := 1 / (x * x)
	r := p[0] + z*(p[1]+z*(p[2]+z*(p[3]+z*(p[4]+z*p[5]))))
	s := 1 + z*(q[0]+z*(q[1]+z*(q[2]+z*(q[3]+z*(q[4]+z*q[5])))))
	return (-0.125 + r/s) / x
}
