// Copyright 2010 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package math32

/*
	Floating-point logarithm of the Gamma function.
*/

// The original C code and the long comment below are
// from FreeBSD's /usr/src/lib/msun/src/e_lgamma_r.c and
// came with this notice.  The go code is a simplified
// version of the original C.
//
// ====================================================
// Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
//
// Developed at SunPro, a Sun Microsystems, Inc. business.
// Permission to use, copy, modify, and distribute this
// software is freely granted, provided that this notice
// is preserved.
// ====================================================
//
// __ieee754_lgamma_r(x, signgamp)
// Reentrant version of the logarithm of the Gamma function
// with user provided pointer for the sign of Gamma(x).
//
// Method:
//   1. Argument Reduction for 0 < x <= 8
//      Since gamma(1+s)=s*gamma(s), for x in [0,8], we may
//      reduce x to a number in [1.5,2.5] by
//              lgamma(1+s) = log(s) + lgamma(s)
//      for example,
//              lgamma(7.3) = log(6.3) + lgamma(6.3)
//                          = log(6.3*5.3) + lgamma(5.3)
//                          = log(6.3*5.3*4.3*3.3*2.3) + lgamma(2.3)
//   2. Polynomial approximation of lgamma around its
//      minimum (ymin=1.461632144968362245) to maintain monotonicity.
//      On [ymin-0.23, ymin+0.27] (i.e., [1.23164,1.73163]), use
//              Let z = x-ymin;
//              lgamma(x) = -1.214862905358496078218 + z**2*poly(z)
//              poly(z) is a 14 degree polynomial.
//   2. Rational approximation in the primary interval [2,3]
//      We use the following approximation:
//              s = x-2.0;
//              lgamma(x) = 0.5*s + s*P(s)/Q(s)
//      with accuracy
//              |P/Q - (lgamma(x)-0.5s)| < 2**-61.71
//      Our algorithms are based on the following observation
//
//                             zeta(2)-1    2    zeta(3)-1    3
// lgamma(2+s) = s*(1-Euler) + --------- * s  -  --------- * s  + ...
//                                 2                 3
//
//      where Euler = 0.5772156649... is the Euler constant, which
//      is very close to 0.5.
//
//   3. For x>=8, we have
//      lgamma(x)~(x-0.5)log(x)-x+0.5*log(2pi)+1/(12x)-1/(360x**3)+....
//      (better formula:
//         lgamma(x)~(x-0.5)*(log(x)-1)-.5*(log(2pi)-1) + ...)
//      Let z = 1/x, then we approximation
//              f(z) = lgamma(x) - (x-0.5)(log(x)-1)
//      by
//                                  3       5             11
//              w = w0 + w1*z + w2*z  + w3*z  + ... + w6*z
//      where
//              |w - f(z)| < 2**-58.74
//
//   4. For negative x, since (G is gamma function)
//              -x*G(-x)*G(x) = pi/sin(pi*x),
//      we have
//              G(x) = pi/(sin(pi*x)*(-x)*G(-x))
//      since G(-x) is positive, sign(G(x)) = sign(sin(pi*x)) for x<0
//      Hence, for x<0, signgam = sign(sin(pi*x)) and
//              lgamma(x) = log(|Gamma(x)|)
//                        = log(pi/(|x*sin(pi*x)|)) - lgamma(-x);
//      Note: one should avoid computing pi*(-x) directly in the
//            computation of sin(pi*(-x)).
//
//   5. Special Cases
//              lgamma(2+s) ~ s*(1-Euler) for tiny s
//              lgamma(1)=lgamma(2)=0
//              lgamma(x) ~ -log(x) for tiny x
//              lgamma(0) = lgamma(inf) = inf
//              lgamma(-integer) = +-inf
//
//

var _lgamA = [...]float32{
	7.7215664089e-02, /* 0x3d9e233f */
	3.2246702909e-01, /* 0x3ea51a66 */
	6.7352302372e-02, /* 0x3d89f001 */
	2.0580807701e-02, /* 0x3ca89915 */
	7.3855509982e-03, /* 0x3bf2027e */
	2.8905137442e-03, /* 0x3b3d6ec6 */
	1.1927076848e-03, /* 0x3a9c54a1 */
	5.1006977446e-04, /* 0x3a05b634 */
	2.2086278477e-04, /* 0x39679767 */
	1.0801156895e-04, /* 0x38e28445 */
	2.5214456400e-05, /* 0x37d383a2 */
	4.4864096708e-05, /* 0x383c2c75 */
}
var _lgamR = [...]float32{
	1.0,              // placeholder
	1.3920053244e+00, /* 0x3fb22d3b */
	7.2193557024e-01, /* 0x3f38d0c5 */
	1.7193385959e-01, /* 0x3e300f6e */
	1.8645919859e-02, /* 0x3c98bf54 */
	7.7794247773e-04, /* 0x3a4beed6 */
	7.3266842264e-06, /* 0x36f5d7bd */
}
var _lgamS = [...]float32{
	-7.7215664089e-02, /* 0xbd9e233f */
	2.1498242021e-01,  /* 0x3e5c245a */
	3.2577878237e-01,  /* 0x3ea6cc7a */
	1.4635047317e-01,  /* 0x3e15dce6 */
	2.6642270386e-02,  /* 0x3cda40e4 */
	1.8402845599e-03,  /* 0x3af135b4 */
	3.1947532989e-05,  /* 0x3805ff67 */
}
var _lgamT = [...]float32{
	4.8383611441e-01,  /* 0x3ef7b95e */
	-1.4758771658e-01, /* 0xbe17213c */
	6.4624942839e-02,  /* 0x3d845a15 */
	-3.2788541168e-02, /* 0xbd064d47 */
	1.7970675603e-02,  /* 0x3c93373d */
	-1.0314224288e-02, /* 0xbc28fcfe */
	6.1005386524e-03,  /* 0x3bc7e707 */
	-3.6845202558e-03, /* 0xbb7177fe */
	2.2596477065e-03,  /* 0x3b141699 */
	-1.4034647029e-03, /* 0xbab7f476 */
	8.8108185446e-04,  /* 0x3a66f867 */
	-5.3859531181e-04, /* 0xba0d3085 */
	3.1563205994e-04,  /* 0x39a57b6b */
	-3.1275415677e-04, /* 0xb9a3f927 */
	3.3552918467e-04,  /* 0x39afe9f7 */
}
var _lgamU = [...]float32{
	-7.7215664089e-02, /* 0xbd9e233f */
	6.3282704353e-01,  /* 0x3f2200f4 */
	1.4549225569e+00,  /* 0x3fba3ae7 */
	9.7771751881e-01,  /* 0x3f7a4bb2 */
	2.2896373272e-01,  /* 0x3e6a7578 */
	1.3381091878e-02,  /* 0x3c5b3c5e */
}
var _lgamV = [...]float32{
	1.0,
	2.4559779167e+00, /* 0x401d2ebe */
	2.1284897327e+00, /* 0x4008392d */
	7.6928514242e-01, /* 0x3f44efdf */
	1.0422264785e-01, /* 0x3dd572af */
	3.2170924824e-03, /* 0x3b52d5db */
}
var _lgamW = [...]float32{
	4.1893854737e-01,  /* 0x3ed67f1d */
	8.3333335817e-02,  /* 0x3daaaaab */
	-2.7777778450e-03, /* 0xbb360b61 */
	7.9365057172e-04,  /* 0x3a500cfd */
	-5.9518753551e-04, /* 0xba1c065c */
	8.3633989561e-04,  /* 0x3a5b3dd2 */
	-1.6309292987e-03, /* 0xbad5c4e8 */
}

// Lgamma returns the natural logarithm and sign (-1 or +1) of Gamma(x).
//
// Special cases are:
//	Lgamma(+Inf) = +Inf
//	Lgamma(0) = +Inf
//	Lgamma(-integer) = +Inf
//	Lgamma(-Inf) = -Inf
//	Lgamma(NaN) = NaN
func Lgamma(x float32) (lgamma float32, sign int) {
	const (
		Ymin  = 1.461632144968362245
		Two23 = 1 << 23                     // 0x4b000000 ~8.388608e+06
		Two58 = 1 << 58                     // 0x5c800000 ~2.8823e+17
		Tiny  = 1.0 / (1 << 70)             // 0x1c800000 ~8.47033e-22
		Tc    = 1.46163214496836224576e+00  // 0x3FF762D86356BE3F
		Tf    = -1.21486290535849611461e-01 // 0xBFBF19B9BCC38A42
		// Tt = -(tail of Tf)
		Tt = -3.63867699703950536541e-18 // 0xBC50C7CAA48A971F
	)
	// special cases
	sign = 1
	switch {
	case IsNaN(x):
		lgamma = x
		return
	case IsInf(x, 0):
		lgamma = x
		return
	case x == 0:
		lgamma = Inf(1)
		return
	}

	neg := false
	if x < 0 {
		x = -x
		neg = true
	}

	if x < Tiny { // if |x| < 2**-70, return -log(|x|)
		if neg {
			sign = -1
		}
		lgamma = -Log(x)
		return
	}
	var nadj float32
	if neg {
		if x >= Two23 { // |x| >= 2**52, must be -integer
			lgamma = Inf(1)
			return
		}
		t := sinPi(x)
		if t == 0 {
			lgamma = Inf(1) // -integer
			return
		}
		nadj = Log(Pi / Abs(t*x))
		if t < 0 {
			sign = -1
		}
	}

	switch {
	case x == 1 || x == 2: // purge off 1 and 2
		lgamma = 0
		return
	case x < 2: // use lgamma(x) = lgamma(x+1) - log(x)
		var y float32
		var i int
		if x <= 0.9 {
			lgamma = -Log(x)
			switch {
			case x >= (Ymin - 1 + 0.27): // 0.7316 <= x <=  0.9
				y = 1 - x
				i = 0
			case x >= (Ymin - 1 - 0.27): // 0.2316 <= x < 0.7316
				y = x - (Tc - 1)
				i = 1
			default: // 0 < x < 0.2316
				y = x
				i = 2
			}
		} else {
			lgamma = 0
			switch {
			case x >= (Ymin + 0.27): // 1.7316 <= x < 2
				y = 2 - x
				i = 0
			case x >= (Ymin - 0.27): // 1.2316 <= x < 1.7316
				y = x - Tc
				i = 1
			default: // 0.9 < x < 1.2316
				y = x - 1
				i = 2
			}
		}
		switch i {
		case 0:
			z := y * y
			p1 := _lgamA[0] + z*(_lgamA[2]+z*(_lgamA[4]+z*(_lgamA[6]+z*(_lgamA[8]+z*_lgamA[10]))))
			p2 := z * (_lgamA[1] + z*(+_lgamA[3]+z*(_lgamA[5]+z*(_lgamA[7]+z*(_lgamA[9]+z*_lgamA[11])))))
			p := y*p1 + p2
			lgamma += (p - 0.5*y)
		case 1:
			z := y * y
			w := z * y
			p1 := _lgamT[0] + w*(_lgamT[3]+w*(_lgamT[6]+w*(_lgamT[9]+w*_lgamT[12]))) // parallel comp
			p2 := _lgamT[1] + w*(_lgamT[4]+w*(_lgamT[7]+w*(_lgamT[10]+w*_lgamT[13])))
			p3 := _lgamT[2] + w*(_lgamT[5]+w*(_lgamT[8]+w*(_lgamT[11]+w*_lgamT[14])))
			p := z*p1 - (Tt - w*(p2+y*p3))
			lgamma += (Tf + p)
		case 2:
			p1 := y * (_lgamU[0] + y*(_lgamU[1]+y*(_lgamU[2]+y*(_lgamU[3]+y*(_lgamU[4]+y*_lgamU[5])))))
			p2 := 1 + y*(_lgamV[1]+y*(_lgamV[2]+y*(_lgamV[3]+y*(_lgamV[4]+y*_lgamV[5]))))
			lgamma += (-0.5*y + p1/p2)
		}
	case x < 8: // 2 <= x < 8
		i := int(x)
		y := x - float32(i)
		p := y * (_lgamS[0] + y*(_lgamS[1]+y*(_lgamS[2]+y*(_lgamS[3]+y*(_lgamS[4]+y*(_lgamS[5]+y*_lgamS[6]))))))
		q := 1 + y*(_lgamR[1]+y*(_lgamR[2]+y*(_lgamR[3]+y*(_lgamR[4]+y*(_lgamR[5]+y*_lgamR[6])))))
		lgamma = 0.5*y + p/q
		z := float32(1.0) // Lgamma(1+s) = Log(s) + Lgamma(s)
		switch i {
		case 7:
			z *= (y + 6)
			fallthrough
		case 6:
			z *= (y + 5)
			fallthrough
		case 5:
			z *= (y + 4)
			fallthrough
		case 4:
			z *= (y + 3)
			fallthrough
		case 3:
			z *= (y + 2)
			lgamma += Log(z)
		}
	case x < Two58: // 8 <= x < 2**58
		t := Log(x)
		z := 1 / x
		y := z * z
		w := _lgamW[0] + z*(_lgamW[1]+y*(_lgamW[2]+y*(_lgamW[3]+y*(_lgamW[4]+y*(_lgamW[5]+y*_lgamW[6])))))
		lgamma = (x-0.5)*(t-1) + w
	default: // 2**58 <= x <= Inf
		lgamma = x * (Log(x) - 1)
	}
	if neg {
		lgamma = nadj - lgamma
	}
	return
}

// sinPi(x) is a helper function for negative x
func sinPi(x float32) float32 {
	const (
		Two23 = 1 << 23 // 0x4b000000 ~8.388608e+06
		Two24 = 1 << 24 // 0x4b800000 ~1.6777216e+07
	)
	if x < 0.25 {
		return -Sin(Pi * x)
	}

	// argument reduction
	z := Floor(x)
	var n int
	if z != x { // inexact
		x = Mod(x, 2)
		n = int(x * 4)
	} else {
		if x >= Two24 { // x must be even
			x = 0
			n = 0
		} else {
			if x < Two23 {
				z = x + Two23 // exact
			}
			n = int(1 & Float32bits(z))
			x = float32(n)
			n <<= 2
		}
	}
	switch n {
	case 0:
		x = Sin(Pi * x)
	case 1, 2:
		x = Cos(Pi * (0.5 - x))
	case 3, 4:
		x = Sin(Pi * (1 - x))
	case 5, 6:
		x = -Cos(Pi * (x - 1.5))
	default:
		x = Sin(Pi * (x - 2))
	}
	return -x
}
