package linux

import (
	"io/ioutil"
	"math"
	"os"
	"strconv"
	"strings"
	"time"

	"github.com/abrander/agento"
)

type MachineStats struct {
	Hostname         string          `json:"h"`
	MemoryStats      *MemoryStats    `json:"m"`
	CpuStats         *CpuStats       `json:"c"`
	DiskStats        *DiskStats      `json:"d"`
	DiskUsageStats   *DiskUsageStats `json:"du"`
	NetStats         *NetStats       `json:"n"`
	LoadStats        *LoadStats      `json:"l"`
	SnmpStats        *SnmpStats      `json:"s"`
	CpuSpeed         *CpuSpeed       `json:"f"`
	AvailableEntropy int64           `json:"e"`
	GatherDuration   time.Duration   `json:"g"`
}

func (m *MachineStats) Gather() {
	start := time.Now()

	m.Hostname, _ = os.Hostname()
	m.MemoryStats = GetMemoryStats()
	m.CpuStats = GetCpuStats()
	m.DiskStats = GetDiskStats()
	m.DiskUsageStats = GetDiskUsageStats()
	m.NetStats = GetNetStats()
	m.LoadStats = GetLoadStats()
	m.SnmpStats = GetSnmpStats()
	m.CpuSpeed = GetCpuSpeed()
	m.AvailableEntropy, _ = getAvailableEntropy()

	m.GatherDuration = time.Now().Sub(start)
}

func (s *MachineStats) GetMap() map[string]interface{} {
	m := make(map[string]interface{})

	s.MemoryStats.GetMap(m)
	s.CpuStats.GetMap(m)
	s.DiskStats.GetMap(m)
	s.DiskUsageStats.GetMap(m)
	s.NetStats.GetMap(m)
	s.LoadStats.GetMap(m)
	s.SnmpStats.GetMap(m)
	s.CpuSpeed.GetMap(m)
	m["misc.AvailableEntropy"] = s.AvailableEntropy

	m["agento.GatherDuration"] = Round(s.GatherDuration.Seconds()*1000.0, 1)

	return m
}

func (s *MachineStats) GetDoc() map[string]string {
	m := make(map[string]string)

	s.MemoryStats.GetDoc(m)
	s.CpuStats.GetDoc(m)
	s.DiskStats.GetDoc(m)
	s.DiskUsageStats.GetDoc(m)
	s.NetStats.GetDoc(m)
	s.LoadStats.GetDoc(m)
	s.SnmpStats.GetDoc(m)
	s.CpuSpeed.GetDoc(m)
	m["misc.AvailableEntropy"] = "Available entropy in the kernel pool (b)"

	m["agento.GatherDuration"] = "Time used to gather metrics for Agento (ms)"

	return m
}

func getAvailableEntropy() (int64, error) {
	contents, err := ioutil.ReadFile("/proc/sys/kernel/random/entropy_avail")

	if err != nil {
		return 0, err
	}

	return strconv.ParseInt(strings.TrimSpace(string(contents)), 10, 64)
}

func Round(value float64, places int) float64 {
	var round float64

	pow := math.Pow(10, float64(places))

	digit := pow * value
	_, div := math.Modf(digit)
	if div >= 0.5 {
		round = math.Ceil(digit)
	} else {
		round = math.Floor(digit)
	}

	return round / pow
}

type MemoryStats struct {
	Used     int64 `json:"u"`
	Free     int64 `json:"f"`
	Shared   int64 `json:"s"`
	Buffers  int64 `json:"b"`
	Cached   int64 `json:"c"`
	SwapUsed int64 `json:"su"`
	SwapFree int64 `json:"sf"`
}

func getMemInfo() *map[string]int64 {
	m := make(map[string]int64)

	path := filepath.Join("/proc/meminfo")
	file, err := os.Open(path)
	if err != nil {
		return &m
	}
	defer file.Close()

	scanner := bufio.NewScanner(file)
	for scanner.Scan() {
		text := scanner.Text()

		n := strings.Index(text, ":")
		if n == -1 {
			continue
		}

		key := text[:n]
		data := strings.Split(strings.Trim(text[(n+1):], " "), " ")
		if len(data) == 1 {
			value, err := strconv.ParseInt(data[0], 10, 64)
			if err != nil {
				continue
			}
			m[key] = value
		} else if len(data) == 2 {
			if data[1] == "kB" {
				value, err := strconv.ParseInt(data[0], 10, 64)
				if err != nil {
					continue
				}

				m[key] = value
			}
		}
	}

	return &m
}

func GetMemoryStats() *MemoryStats {
	stat := MemoryStats{}
	meminfo := getMemInfo()

	stat.Used = (*meminfo)["MemTotal"] - (*meminfo)["MemFree"] - (*meminfo)["Buffers"] - (*meminfo)["Cached"]
	stat.Free = (*meminfo)["MemFree"]
	stat.Shared = (*meminfo)["Shmem"]
	stat.Buffers = (*meminfo)["Buffers"]
	stat.Cached = (*meminfo)["Cached"]

	stat.SwapUsed = (*meminfo)["SwapTotal"] - (*meminfo)["SwapFree"]
	stat.SwapFree = (*meminfo)["SwapFree"]

	return &stat
}

func (s *MemoryStats) GetMap(m map[string]interface{}) {
	m["mem.Used"] = s.Used
	m["mem.Free"] = s.Free
	m["mem.Shared"] = s.Shared
	m["mem.Buffers"] = s.Buffers
	m["mem.Cached"] = s.Cached
	m["swap.Used"] = s.SwapUsed
	m["swap.Free"] = s.SwapFree
}

func (s *MemoryStats) GetDoc(m map[string]string) {
	m["mem.Used"] = "Memory used (b)"
	m["mem.Free"] = "Free memory (b)"
	m["mem.Shared"] = "Memory shared among multiple processes (b)"
	m["mem.Buffers"] = "Memory used for buffers (b)"
	m["mem.Cached"] = "Memory used for cache (b)"
	m["swap.Used"] = "Used swap (b)"
	m["swap.Free"] = "Free swap (b)"
}
